import os
import pytest
import icalendar
import pytz
from datetime import datetime
from dateutil import tz
try:
    import zoneinfo
except ModuleNotFoundError:
    from backports import zoneinfo

class DataSource:
    '''A collection of parsed ICS elements (e.g calendars, timezones, events)'''
    def __init__(self, data_source_folder, parser):
        self._parser = parser
        self._data_source_folder = data_source_folder

    def keys(self):
        """Return all the files that could be used."""
        return [file[:-4] for file in os.listdir(self._data_source_folder) if file.lower().endswith(".ics")]

    def __getattr__(self, attribute):
        """Parse a file and return the result stored in the attribute."""
        source_file = attribute.replace('-', '_') + '.ics'
        source_path = os.path.join(self._data_source_folder, source_file)
        if not os.path.isfile(source_path):
            raise AttributeError(f"{source_path} does not exist.")
        with open(source_path, 'rb') as f:
            raw_ics = f.read()
        source = self._parser(raw_ics)
        if not isinstance(source, list):
            source.raw_ics = raw_ics
        self.__dict__[attribute] = source
        return source

    def __getitem__(self, key):
        return getattr(self, key)

    def __repr__(self):
        return repr(self.__dict__)

    @property
    def multiple(self):
        """Return a list of all components parsed."""
        return self.__class__(self._data_source_folder, lambda data: self._parser(data, multiple=True))

HERE = os.path.dirname(__file__)
CALENDARS_FOLDER = os.path.join(HERE, 'calendars')
CALENDARS = DataSource(CALENDARS_FOLDER, icalendar.Calendar.from_ical)
TIMEZONES_FOLDER = os.path.join(HERE, 'timezones')
TIMEZONES = DataSource(TIMEZONES_FOLDER, icalendar.Timezone.from_ical)
EVENTS_FOLDER = os.path.join(HERE, 'events')
EVENTS = DataSource(EVENTS_FOLDER, icalendar.Event.from_ical)

@pytest.fixture()
def calendars():
    return CALENDARS

@pytest.fixture()
def timezones():
    return TIMEZONES

@pytest.fixture()
def events():
    return EVENTS

@pytest.fixture(params=[
    pytz.utc,
    zoneinfo.ZoneInfo('UTC'),
    pytz.timezone('UTC'),
    tz.UTC,
    tz.gettz('UTC')])
def utc(request):
    return request.param

@pytest.fixture(params=[
    lambda dt, tzname: pytz.timezone(tzname).localize(dt),
    lambda dt, tzname: dt.replace(tzinfo=tz.gettz(tzname)),
    lambda dt, tzname: dt.replace(tzinfo=zoneinfo.ZoneInfo(tzname))
])
def in_timezone(request):
    return request.param


@pytest.fixture(params=[
    (data, key)
    for data in [CALENDARS, TIMEZONES, EVENTS]
    for key in data.keys() if key not in
    ( # exclude broken calendars here
        "big_bad_calendar", "issue_104_broken_calendar", "small_bad_calendar",
        "multiple_calendar_components", "pr_480_summary_with_colon"
    )
])
def ics_file(request):
    """An example ICS file."""
    data, key = request.param
    print(key)
    return data[key]



# icalendar documentation build configuration file
import pkg_resources
import datetime
import os

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

try:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except ImportError:
    html_theme = 'default'
    if not on_rtd:
        print('-' * 74)
        print('Warning: sphinx-rtd-theme not installed, building with default '
              'theme.')
        print('-' * 74)

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode'
]
source_suffix = '.rst'
master_doc = 'index'

project = 'icalendar'
this_year = datetime.date.today().year
copyright = f'{this_year}, Plone Foundation'
version = pkg_resources.get_distribution('icalendar').version
release = version

exclude_patterns = ['_build', 'lib', 'bin', 'include', 'local']
pygments_style = 'sphinx'

htmlhelp_basename = 'icalendardoc'

man_pages = [
    ('index', 'icalendar', 'icalendar Documentation',
     ['Plone Foundation'], 1)
]



"""Test the equality and inequality of components."""
import copy
import pytz
from icalendar.prop import *
from datetime import datetime, date, timedelta
import pytest


def test_parsed_calendars_are_equal_if_parsed_again(ics_file):
    """Ensure that a calendar equals the same calendar."""
    copy_of_calendar = ics_file.__class__.from_ical(ics_file.to_ical())
    assert copy_of_calendar == ics_file
    assert not copy_of_calendar != ics_file


def test_parsed_calendars_are_equal_if_from_same_source(ics_file):
    """Ensure that a calendar equals the same calendar."""
    cal1 = ics_file.__class__.from_ical(ics_file.raw_ics)
    cal2 = ics_file.__class__.from_ical(ics_file.raw_ics)
    assert cal1 == cal2
    assert not cal1 != cal2


def test_copies_are_equal(ics_file):
    """Ensure that copies are equal."""
    copy1 = ics_file.copy(); copy1.subcomponents = ics_file.subcomponents
    copy2 = ics_file.copy();  copy2.subcomponents = ics_file.subcomponents[:]
    assert copy1 == copy2
    assert copy1 == ics_file
    assert copy2 == ics_file
    assert not copy1 != copy2
    assert not copy1 != ics_file
    assert not copy2 != ics_file


def test_copy_does_not_copy_subcomponents(calendars):
    """If we copy the subcomponents, assumptions around copies will be broken."""
    assert calendars.timezoned.subcomponents
    assert not calendars.timezoned.copy().subcomponents


def test_deep_copies_are_equal(ics_file):
    """Ensure that deep copies are equal."""
    try:
        assert copy.deepcopy(ics_file) == copy.deepcopy(ics_file)
        assert copy.deepcopy(ics_file) == ics_file
        assert not copy.deepcopy(ics_file) != copy.deepcopy(ics_file)
        assert not copy.deepcopy(ics_file) != ics_file
    except pytz.UnknownTimeZoneError:
        # Ignore errors when a custom time zone is used.
        # This is still covered by the parsing test.
        pass


def test_vGeo():
    """Check the equality of vGeo."""
    assert vGeo(("100", "12.33")) == vGeo(("100.00", "12.330"))
    assert vGeo(("100", "12.331")) != vGeo(("100.00", "12.330"))
    assert vGeo(("10", "12.33")) != vGeo(("100.00", "12.330"))


def test_vBinary():
    assert vBinary('asd') == vBinary('asd')
    assert vBinary('asdf') != vBinary('asd')


def test_vBoolean():
    assert vBoolean.from_ical('TRUE') == vBoolean.from_ical('TRUE')
    assert vBoolean.from_ical('FALSE') == vBoolean.from_ical('FALSE')
    assert vBoolean.from_ical('TRUE') != vBoolean.from_ical('FALSE')


def test_vCategory():
    assert vCategory("HELLO") == vCategory("HELLO")
    assert vCategory(["a","b"]) == vCategory(["a","b"])
    assert vCategory(["a","b"]) != vCategory(["a","b", "c"])


def test_vText():
    assert vText("HELLO") == vText("HELLO")
    assert not vText("HELLO") != vText("HELLO")
    assert vText("HELLO1") != vText("HELLO")
    assert not vText("HELLO1") == vText("HELLO")


@pytest.mark.parametrize(
    "vType,v1,v2",
    [
        (vDatetime, datetime(2023, 11, 1, 10, 11), datetime(2023, 11, 1, 10, 10)),
        (vDate, date(2023, 11, 1), date(2023, 10, 31)),
        (vDuration, timedelta(3, 11, 1), timedelta(23, 10, 31)),
        (vPeriod, (datetime(2023, 11, 1, 10, 11), timedelta(3, 11, 1)), (datetime(2023, 11, 1, 10, 11), timedelta(23, 10, 31))),
        (vPeriod, (datetime(2023, 11, 1, 10, 1), timedelta(3, 11, 1)), (datetime(2023, 11, 1, 10, 11), timedelta(3, 11, 1))),
        (vPeriod, (datetime(2023, 11, 1, 10, 1), datetime(2023, 11, 1, 10, 3)), (datetime(2023, 11, 1, 10, 1), datetime(2023, 11, 1, 10, 2))),
        (vTime, time(10, 10, 10), time(10, 10, 11)),
    ]
)
@pytest.mark.parametrize("eq", ["==", "!="])
@pytest.mark.parametrize("cls1", [0, 1])
@pytest.mark.parametrize("cls2", [0, 1])
@pytest.mark.parametrize("hash", [lambda x:x, hash])
def test_vDDDTypes_and_others(vType, v1, v2, cls1, cls2, eq, hash):
    """Check equality and inequality."""
    t1 = (vType, vDDDTypes)[cls1]
    t2 = (vType, vDDDTypes)[cls2]
    if eq == "==":
        assert hash(v1) == hash(v1)
        assert hash(t1(v1)) == hash(t2(v1))
        assert not hash(t1(v1)) != hash(t2(v1))
    else:
        assert hash(v1) != hash(v2)
        assert hash(t1(v1)) != hash(t2(v2))


def test_repr_vDDDTypes():
    assert "vDDDTypes" in repr(vDDDTypes(timedelta(3, 11, 1)))


vDDDLists_examples = [
    vDDDLists([]),
    vDDDLists([datetime(2023, 11, 1, 10, 1)]),
    vDDDLists([datetime(2023, 11, 1, 10, 1), date(2023, 11, 1)]),
]
@pytest.mark.parametrize("l1", vDDDLists_examples)
@pytest.mark.parametrize("l2", vDDDLists_examples)
def test_vDDDLists(l1, l2):
    """Check the equality functions of vDDDLists."""
    equal = l1 is l2
    l2 = copy.deepcopy(l2)
    assert equal == (l1 == l2)
    assert equal != (l1 != l2)



import itertools
from datetime import datetime
from datetime import timedelta
import unittest

import pytest

import icalendar
import pytz
import re


class TestCalComponent(unittest.TestCase):

    def test_cal_Component(self):
        from icalendar.cal import Component, Calendar, Event
        from icalendar import prop

        # A component is like a dictionary with extra methods and attributes.
        c = Component()
        c.name = 'VCALENDAR'

        self.assertTrue(c)
        self.assertTrue(c.is_empty())

        # Every key defines a property.A property can consist of either a
        # single item. This can be set with a single value...
        c['prodid'] = '-//max m//icalendar.mxm.dk/'
        self.assertFalse(c.is_empty())

        self.assertEqual(
            c,
            Calendar({'PRODID': '-//max m//icalendar.mxm.dk/'})
        )

        # or with a list
        c['ATTENDEE'] = ['Max M', 'Rasmussen']
        self.assertEqual(
            c,
            Calendar({'ATTENDEE': ['Max M', 'Rasmussen'],
                      'PRODID': '-//max m//icalendar.mxm.dk/'})
        )

        # ## ADD MULTIPLE VALUES TO A PROPERTY

        # if you use the add method you don't have to considder if a value is
        # a list or not.
        c = Component()
        c.name = 'VEVENT'

        # add multiple values at once
        c.add('attendee',
              ['test@test.com', 'test2@test.com'])

        # or add one per line
        c.add('attendee', 'maxm@mxm.dk')
        c.add('attendee', 'test@example.dk')

        # add again multiple values at once to very concatenaton of lists
        c.add('attendee',
              ['test3@test.com', 'test4@test.com'])

        self.assertEqual(
            c,
            Event({'ATTENDEE': [
                prop.vCalAddress('test@test.com'),
                prop.vCalAddress('test2@test.com'),
                prop.vCalAddress('maxm@mxm.dk'),
                prop.vCalAddress('test@example.dk'),
                prop.vCalAddress('test3@test.com'),
                prop.vCalAddress('test4@test.com')
            ]})
        )

        ###

        # You can get the values back directly ...
        c.add('prodid', '-//my product//')
        self.assertEqual(c['prodid'], prop.vText('-//my product//'))

        # ... or decoded to a python type
        self.assertEqual(c.decoded('prodid'), b'-//my product//')

        # With default values for non existing properties
        self.assertEqual(c.decoded('version', 'No Version'), 'No Version')

        c.add('rdate', [datetime(2013, 3, 28), datetime(2013, 3, 27)])
        self.assertTrue(isinstance(c.decoded('rdate'), prop.vDDDLists))

        # The component can render itself in the RFC 2445 format.
        c = Component()
        c.name = 'VCALENDAR'
        c.add('attendee', 'Max M')
        self.assertEqual(
            c.to_ical(),
            b'BEGIN:VCALENDAR\r\nATTENDEE:Max M\r\nEND:VCALENDAR\r\n'
        )

        # Components can be nested, so You can add a subcomponent. Eg a calendar
        # holds events.
        e = Component(summary='A brief history of time')
        e.name = 'VEVENT'
        e.add('dtend', '20000102T000000', encode=0)
        e.add('dtstart', '20000101T000000', encode=0)
        self.assertEqual(
            e.to_ical(),
            b'BEGIN:VEVENT\r\nDTEND:20000102T000000\r\n'
            + b'DTSTART:20000101T000000\r\nSUMMARY:A brief history of time\r'
            + b'\nEND:VEVENT\r\n'
        )

        c.add_component(e)
        self.assertEqual(
            c.subcomponents,
            [Event({'DTEND': '20000102T000000', 'DTSTART': '20000101T000000',
                    'SUMMARY': 'A brief history of time'})]
        )

        # We can walk over nested componentes with the walk method.
        self.assertEqual([i.name for i in c.walk()], ['VCALENDAR', 'VEVENT'])

        # We can also just walk over specific component types, by filtering
        # them on their name.
        self.assertEqual([i.name for i in c.walk('VEVENT')], ['VEVENT'])

        self.assertEqual(
            [i['dtstart'] for i in c.walk('VEVENT')],
            ['20000101T000000']
        )

        # We can enumerate property items recursively with the property_items
        # method.
        self.assertEqual(
            c.property_items(),
            [('BEGIN', b'VCALENDAR'), ('ATTENDEE', prop.vCalAddress('Max M')),
             ('BEGIN', b'VEVENT'), ('DTEND', '20000102T000000'),
             ('DTSTART', '20000101T000000'),
             ('SUMMARY', 'A brief history of time'), ('END', b'VEVENT'),
             ('END', b'VCALENDAR')]
        )

        # We can also enumerate property items just under the component.
        self.assertEqual(
            c.property_items(recursive=False),
            [('BEGIN', b'VCALENDAR'),
             ('ATTENDEE', prop.vCalAddress('Max M')),
             ('END', b'VCALENDAR')]
        )

        sc = c.subcomponents[0]
        self.assertEqual(
            sc.property_items(recursive=False),
            [('BEGIN', b'VEVENT'), ('DTEND', '20000102T000000'),
             ('DTSTART', '20000101T000000'),
             ('SUMMARY', 'A brief history of time'), ('END', b'VEVENT')]
        )

        # Text fields which span multiple mulitple lines require proper
        # indenting
        c = Calendar()
        c['description'] = 'Paragraph one\n\nParagraph two'
        self.assertEqual(
            c.to_ical(),
            b'BEGIN:VCALENDAR\r\nDESCRIPTION:Paragraph one\\n\\nParagraph two'
            + b'\r\nEND:VCALENDAR\r\n'
        )

        # INLINE properties have their values on one property line. Note the
        # double quoting of the value with a colon in it.
        c = Calendar()
        c['resources'] = 'Chair, Table, "Room: 42"'
        self.assertEqual(
            c,
            Calendar({'RESOURCES': 'Chair, Table, "Room: 42"'})
        )

        self.assertEqual(
            c.to_ical(),
            b'BEGIN:VCALENDAR\r\nRESOURCES:Chair\\, Table\\, "Room: 42"\r\n'
            + b'END:VCALENDAR\r\n'
        )

        # The inline values must be handled by the get_inline() and
        # set_inline() methods.
        self.assertEqual(
            c.get_inline('resources', decode=0),
            ['Chair', 'Table', 'Room: 42']
        )

        # These can also be decoded
        self.assertEqual(
            c.get_inline('resources', decode=1),
            [b'Chair', b'Table', b'Room: 42']
        )

        # You can set them directly ...
        c.set_inline('resources', ['A', 'List', 'of', 'some, recources'],
                     encode=1)
        self.assertEqual(c['resources'], 'A,List,of,"some, recources"')

        # ... and back again
        self.assertEqual(
            c.get_inline('resources', decode=0),
            ['A', 'List', 'of', 'some, recources']
        )

        c['freebusy'] = '19970308T160000Z/PT3H,19970308T200000Z/PT1H,'\
                        + '19970308T230000Z/19970309T000000Z'
        self.assertEqual(
            c.get_inline('freebusy', decode=0),
            ['19970308T160000Z/PT3H', '19970308T200000Z/PT1H',
             '19970308T230000Z/19970309T000000Z']
        )

        freebusy = c.get_inline('freebusy', decode=1)
        self.assertTrue(isinstance(freebusy[0][0], datetime))
        self.assertTrue(isinstance(freebusy[0][1], timedelta))

    def test_cal_Component_add(self):
        # Test the for timezone correctness: dtstart should preserve it's
        # timezone, created, dtstamp and last-modified must be in UTC.
        Component = icalendar.cal.Component
        comp = Component()
        vienna = pytz.timezone("Europe/Vienna")
        comp.add('dtstart', vienna.localize(datetime(2010, 10, 10, 10, 0, 0)))
        comp.add('created', datetime(2010, 10, 10, 12, 0, 0))
        comp.add('dtstamp', vienna.localize(datetime(2010, 10, 10, 14, 0, 0)))
        comp.add('last-modified', pytz.utc.localize(
            datetime(2010, 10, 10, 16, 0, 0)))

        lines = comp.to_ical().splitlines()
        self.assertTrue(
            b"DTSTART;TZID=Europe/Vienna:20101010T100000"
            in lines)
        self.assertTrue(b"CREATED:20101010T120000Z" in lines)
        self.assertTrue(b"DTSTAMP:20101010T120000Z" in lines)
        self.assertTrue(
            b"LAST-MODIFIED:20101010T160000Z" in lines
        )

    def test_cal_Component_add_no_reencode(self):
        """Already encoded values should not be re-encoded.
        """
        from icalendar import cal, prop
        comp = cal.Component()
        comp.add('ATTACH', 'me')

        comp.add('ATTACH', 'you', encode=False)
        binary = prop.vBinary('us')
        comp.add('ATTACH', binary)

        self.assertEqual(comp['ATTACH'], ['me', 'you', binary])

    def test_cal_Component_add_property_parameter(self):
        # Test the for timezone correctness: dtstart should preserve it's
        # timezone, crated, dtstamp and last-modified must be in UTC.
        Component = icalendar.cal.Component
        comp = Component()
        comp.add('X-TEST-PROP', 'tryout.',
                 parameters={'prop1': 'val1', 'prop2': 'val2'})
        lines = comp.to_ical().splitlines()
        self.assertTrue(b"X-TEST-PROP;PROP1=val1;PROP2=val2:tryout." in lines)

    def test_cal_Component_from_ical(self):
        # Check for proper handling of TZID parameter of datetime properties
        Component = icalendar.cal.Component
        for component_name, property_name in (
            ('VEVENT', 'DTSTART'),
            ('VEVENT', 'DTEND'),
            ('VEVENT', 'RECURRENCE-ID'),
            ('VTODO', 'DUE')
        ):
            component_str = 'BEGIN:' + component_name + '\n'
            component_str += property_name + ';TZID=America/Denver:'
            component_str += '20120404T073000\nEND:' + component_name
            component = Component.from_ical(component_str)
            self.assertEqual(str(component[property_name].dt.tzinfo.zone),
                             "America/Denver")

            component_str = 'BEGIN:' + component_name + '\n'
            component_str += property_name + ':'
            component_str += '20120404T073000\nEND:' + component_name
            component = Component.from_ical(component_str)
            self.assertEqual(component[property_name].dt.tzinfo,
                             None)

    def test_cal_Component_to_ical_property_order(self):
        Component = icalendar.cal.Component
        component_str = [b'BEGIN:VEVENT',
                         b'DTSTART:19970714T170000Z',
                         b'DTEND:19970715T035959Z',
                         b'SUMMARY:Bastille Day Party',
                         b'END:VEVENT']
        component = Component.from_ical(b'\r\n'.join(component_str))

        sorted_str = component.to_ical().splitlines()
        assert sorted_str != component_str
        assert set(sorted_str) == set(component_str)

        preserved_str = component.to_ical(sorted=False).splitlines()
        assert preserved_str == component_str

    def test_cal_Component_to_ical_parameter_order(self):
        Component = icalendar.cal.Component
        component_str = [b'BEGIN:VEVENT',
                         b'X-FOOBAR;C=one;A=two;B=three:helloworld.',
                         b'END:VEVENT']
        component = Component.from_ical(b'\r\n'.join(component_str))

        sorted_str = component.to_ical().splitlines()
        assert sorted_str[0] == component_str[0]
        assert sorted_str[1] == b'X-FOOBAR;A=two;B=three;C=one:helloworld.'
        assert sorted_str[2] == component_str[2]

        preserved_str = component.to_ical(sorted=False).splitlines()
        assert preserved_str == component_str

    def test_repr(self):
        """Test correct class representation.
        """
        from icalendar.cal import Component, Calendar, Event

        component = Component()
        component['key1'] = 'value1'

        self.assertTrue(
            re.match(r"Component\({u?'KEY1': u?'value1'}\)", str(component))
        )

        calendar = Calendar()
        calendar['key1'] = 'value1'

        self.assertTrue(
            re.match(r"VCALENDAR\({u?'KEY1': u?'value1'}\)", str(calendar))
        )

        event = Event()
        event['key1'] = 'value1'

        self.assertTrue(
            re.match(r"VEVENT\({u?'KEY1': u?'value1'}\)", str(event))
        )

        # Representation of nested Components
        nested = Component(key1='VALUE1')
        nested.add_component(component)
        calendar.add_component(event)
        nested.add_component(calendar)

        self.assertTrue(
            re.match(
                r"Component\({u?'KEY1': u?'VALUE1'}, "
                r"Component\({u?'KEY1': u?'value1'}\), "
                r"VCALENDAR\({u?'KEY1': u?'value1'}, "
                r"VEVENT\({u?'KEY1': u?'value1'}\)\)\)",
                str(nested)
            )
        )


class TestCal(unittest.TestCase):

    def test_cal_ComponentFactory(self):
        ComponentFactory = icalendar.cal.ComponentFactory
        factory = ComponentFactory()
        component = factory['VEVENT']
        event = component(dtstart='19700101')
        self.assertEqual(
            event.to_ical(),
            b'BEGIN:VEVENT\r\nDTSTART:19700101\r\nEND:VEVENT\r\n'
        )

        self.assertEqual(
            factory.get('VCALENDAR', icalendar.cal.Component),
            icalendar.cal.Calendar)

    def test_cal_Calendar(self):
        # Setting up a minimal calendar component looks like this
        cal = icalendar.cal.Calendar()

        # Some properties are required to be compliant
        cal['prodid'] = '-//My calendar product//mxm.dk//'
        cal['version'] = '2.0'

        # We also need at least one subcomponent for a calendar to be compliant
        event = icalendar.cal.Event()
        event['summary'] = 'Python meeting about calendaring'
        event['uid'] = '42'
        event.add('dtstart', datetime(2005, 4, 4, 8, 0, 0))
        cal.add_component(event)
        self.assertEqual(
            cal.subcomponents[0].to_ical(),
            b'BEGIN:VEVENT\r\nSUMMARY:Python meeting about calendaring\r\n'
            + b'DTSTART:20050404T080000\r\nUID:42\r\n'
            + b'END:VEVENT\r\n')

        # Write to disc
        import tempfile
        import os
        directory = tempfile.mkdtemp()
        with open(os.path.join(directory, 'test.ics'), 'wb') as fp:
            fp.write(cal.to_ical())

        # Parsing a complete calendar from a string will silently ignore wrong
        # events but adding the error information to the component's 'errors'
        # attribute. The error in the following is the third EXDATE: it has an
        # empty DATE.
        s = '\r\n'.join(('BEGIN:VCALENDAR',
                         'PRODID:-//Google Inc//Google Calendar 70.9054//EN',
                         'VERSION:2.0',
                         'CALSCALE:GREGORIAN',
                         'METHOD:PUBLISH',
                         'BEGIN:VEVENT',
                         'DESCRIPTION:Perfectly OK event',
                         'DTSTART;VALUE=DATE:20080303',
                         'DTEND;VALUE=DATE:20080304',
                         'RRULE:FREQ=DAILY;UNTIL=20080323T235959Z',
                         'EXDATE;VALUE=DATE:20080311',
                         'END:VEVENT',
                         'BEGIN:VEVENT',
                         'DESCRIPTION:Wrong event',
                         'DTSTART;VALUE=DATE:20080303',
                         'DTEND;VALUE=DATE:20080304',
                         'RRULE:FREQ=DAILY;UNTIL=20080323T235959Z',
                         'EXDATE;VALUE=DATE:20080311',
                         'EXDATE;VALUE=DATE:',
                         'END:VEVENT',
                         'END:VCALENDAR'))
        self.assertEqual(
            [e['DESCRIPTION'].to_ical()
                for e in icalendar.cal.Calendar.from_ical(s).walk('VEVENT')],
            [b'Perfectly OK event', b'Wrong event'])
        self.assertEqual(
            [e.errors
                for e in icalendar.cal.Calendar.from_ical(s).walk('VEVENT')],
            [[], [('EXDATE', "Expected datetime, date, or time, got: ''")]]
        )

    def test_cal_strict_parsing(self):
        cal_str = b'\r\n'.join(
            [
                b'BEGIN:VCALENDAR',
                b'BEGIN:VTIMEZONE',
                b'TZID:Europe/Prague',
                b'BEGIN:STANDARD',
                b'DTSTART:18500101T000000',
                b'TZNAME:PMT',
                b'TZOFFSETFROM:+5744',
                b'TZOFFSETTO:+5744',
                b'END:STANDARD',
                b'END:VTIMEZONE',
                b'END:VCALENDAR',
                b'',
            ]
        )

        self.assertRaises(ValueError, icalendar.Calendar.from_ical, cal_str)
        icalendar.vUTCOffset.ignore_exceptions = True
        self.assertEqual(icalendar.Calendar.from_ical(cal_str).to_ical(), cal_str)
        icalendar.vUTCOffset.ignore_exceptions = False


@pytest.mark.parametrize(
    'calendar, other_calendar',
    itertools.product([
        'issue_156_RDATE_with_PERIOD_TZID_khal',
        'issue_156_RDATE_with_PERIOD_TZID_khal_2',
        'issue_178_custom_component_contains_other',
        'issue_178_custom_component_inside_other',
        'issue_526_calendar_with_events',
        'issue_526_calendar_with_different_events',
        'issue_526_calendar_with_event_subset',
    ], repeat=2)
)
def test_comparing_calendars(calendars, calendar, other_calendar):
    are_calendars_equal = calendars[calendar] == calendars[other_calendar]
    are_calendars_actually_equal = calendar == other_calendar
    assert are_calendars_equal == are_calendars_actually_equal


@pytest.mark.parametrize('calendar, shuffeled_calendar', [
    (
        'issue_526_calendar_with_events',
        'issue_526_calendar_with_shuffeled_events',
    ),
])
def test_calendars_with_same_subcomponents_in_different_order_are_equal(calendars, calendar, shuffeled_calendar):
    assert not calendars[calendar].subcomponents == calendars[shuffeled_calendar].subcomponents
    assert calendars[calendar] == calendars[shuffeled_calendar]




"""These tests make sure that we have some coverage on the usage of the PERIOD value type.

See
- https://github.com/collective/icalendar/issues/156
- https://github.com/pimutils/khal/issues/152#issuecomment-933635248
"""
import pytest
import pytz
from icalendar.prop import vDDDTypes
import datetime


@pytest.mark.parametrize("calname,tzname,index,period_string", [
    ("issue_156_RDATE_with_PERIOD_TZID_khal_2", "Europe/Berlin", 0, "20211101T160000/20211101T163000"),
    ("issue_156_RDATE_with_PERIOD_TZID_khal_2", "Europe/Berlin", 1, "20211206T160000/20211206T163000"),
    ("issue_156_RDATE_with_PERIOD_TZID_khal_2", "Europe/Berlin", 2, "20220103T160000/20220103T163000"),
    ("issue_156_RDATE_with_PERIOD_TZID_khal_2", "Europe/Berlin", 3, "20220207T160000/20220207T163000"),
] + [
    ("issue_156_RDATE_with_PERIOD_TZID_khal", "America/Chicago", i, period)
    for i, period in enumerate(("20180327T080000/20180327T0"
        "90000,20180403T080000/20180403T090000,20180410T080000/20180410T090000,2018"
        "0417T080000/20180417T090000,20180424T080000/20180424T090000,20180501T08000"
        "0/20180501T090000,20180508T080000/20180508T090000,20180515T080000/20180515"
        "T090000,20180522T080000/20180522T090000,20180529T080000/20180529T090000,20"
        "180605T080000/20180605T090000,20180612T080000/20180612T090000,20180619T080"
        "000/20180619T090000,20180626T080000/20180626T090000,20180703T080000/201807"
        "03T090000,20180710T080000/20180710T090000,20180717T080000/20180717T090000,"
        "20180724T080000/20180724T090000,20180731T080000/20180731T090000").split(","))
])
def test_issue_156_period_list_in_rdate(calendars, calname, tzname, index, period_string):
    """Check items in a list of period values."""
    calendar = calendars[calname]
    rdate = calendar.walk("vevent")[0]["rdate"]
    period = rdate.dts[index]
    assert period.dt == vDDDTypes.from_ical(period_string, timezone=tzname)


def test_duration_properly_parsed(events):
    """This checks the duration PT5H30M."""
    start = vDDDTypes.from_ical("19970109T180000Z")
    duration = vDDDTypes.from_ical("PT5H30M")
    rdate = events.issue_156_RDATE_with_PERIOD_list["RDATE"]
    print(rdate)
    period = rdate.dts[1].dt
    print(dir(duration))
    assert period[0] == start
    assert period[1].days == 0
    assert period[1].seconds == (5 * 60 + 30) * 60
    assert period[1] == duration


def test_tzid_is_part_of_the_parameters(calendars):
    """The TZID should be mentioned in the parameters."""
    event = list(calendars.period_with_timezone.walk("VEVENT"))[0]
    assert event["RDATE"].params["TZID"] == "America/Vancouver"


def test_tzid_is_part_of_the_period_values(calendars):
    """The TZID should be set in the datetime."""
    event = list(calendars.period_with_timezone.walk("VEVENT"))[0]
    start, end = event["RDATE"].dts[0].dt
    assert start == pytz.timezone("America/Vancouver").localize(datetime.datetime(2023, 12, 13, 12))
    assert end == pytz.timezone("America/Vancouver").localize(datetime.datetime(2023, 12, 13, 15))



from icalendar import Calendar, Event


def test_issue_322_single_string_split_into_multiple_categories(calendars):
    calendar = Calendar()
    event = Event()
    event.add('summary', 'Event with bare string as argument for categories')
    event.add('categories', "Lecture")
    calendar.add_component(event)
    assert calendar.to_ical() == calendars.issue_322_expected_calendar.raw_ics



'''tests ensuring that *the* way of doing things works'''

import datetime
from icalendar import Calendar, Event
import pytest


def test_creating_calendar_with_unicode_fields(calendars, utc):
    ''' create a calendar with events that contain unicode characters in their fields '''
    cal = Calendar()
    cal.add('PRODID', '-//Plönë.org//NONSGML plone.app.event//EN')
    cal.add('VERSION', '2.0')
    cal.add('X-WR-CALNAME', 'äöü ÄÖÜ €')
    cal.add('X-WR-CALDESC', 'test non ascii: äöü ÄÖÜ €')
    cal.add('X-WR-RELCALID', '12345')

    event = Event()
    event.add('DTSTART', datetime.datetime(2010, 10, 10, 10, 0, 0, tzinfo=utc))
    event.add('DTEND', datetime.datetime(2010, 10, 10, 12, 0, 0, tzinfo=utc))
    event.add('CREATED', datetime.datetime(2010, 10, 10, 0, 0, 0, tzinfo=utc))
    event.add('UID', '123456')
    event.add('SUMMARY', 'Non-ASCII Test: ÄÖÜ äöü €')
    event.add('DESCRIPTION', 'icalendar should be able to de/serialize non-ascii.')
    event.add('LOCATION', 'Tribstrül')
    cal.add_component(event)

    # test_create_event_simple
    event1 = Event()
    event1.add('DTSTART', datetime.datetime(2010, 10, 10, 0, 0, 0, tzinfo=utc))
    event1.add('SUMMARY', 'åäö')
    cal.add_component(event1)

    # test_unicode_parameter_name
    # test for issue #80 https://github.com/collective/icalendar/issues/80
    event2 = Event()
    event2.add('DESCRIPTION', 'äöüßÄÖÜ')
    cal.add_component(event2)

    assert cal.to_ical() == calendars.created_calendar_with_unicode_fields.raw_ics



'''Issue #27 - multiple periods

   https://github.com/collective/icalendar/issues/27
'''
from icalendar import Calendar

def test_issue_27_multiple_periods(calendars):
    free_busy = list(calendars.issue_27_multiple_periods_in_freebusy_multiple_freebusies.walk('VFREEBUSY'))[0]
    free_busy_period = free_busy['freebusy']
    print(free_busy['freebusy'])
    equivalent_way_of_defining_free_busy = list(calendars.issue_27_multiple_periods_in_freebusy_one_freebusy.walk('VFREEBUSY'))[0]
    free_busy_period_equivalent = equivalent_way_of_defining_free_busy['freebusy']
    assert free_busy_period == free_busy_period_equivalent




import pytest
import unittest

from icalendar.tools import UIDGenerator

class TestTools(unittest.TestCase):

    def test_tools_UIDGenerator(self):

        # Automatic semi-random uid
        g = UIDGenerator()
        uid = g.uid()

        txt = uid.to_ical()
        length = 15 + 1 + 16 + 1 + 11
        self.assertTrue(len(txt) == length)
        self.assertTrue(b'@example.com' in txt)

        # You should at least insert your own hostname to be more compliant
        uid = g.uid('Example.ORG')
        txt = uid.to_ical()
        self.assertTrue(len(txt) == length)
        self.assertTrue(b'@Example.ORG' in txt)

        # You can also insert a path or similar
        uid = g.uid('Example.ORG', '/path/to/content')
        txt = uid.to_ical()
        self.assertTrue(len(txt) == length)
        self.assertTrue(b'-/path/to/content@Example.ORG' in txt)


@pytest.mark.parametrize('split,expected,args,kw', [
    # default argument host_name
    ("@", "example.com", (), {},),
    ("@", "example.com", ("example.com",), {}),
    ("@", "example.com", (), {"host_name":"example.com"}),
    # replaced host_name
    ("@", "test.test", ("test.test",), {}),
    ("@", "test.test", (), {"host_name":"test.test"}),
    # replace unique
    ("-", "123@example.com", (), {"unique": "123"},),
    ("-", "abc@example.com", (), {"unique": "abc"},),
    # replace host_name and unique
    ("-", "1234@test.icalendar", (), {"unique": "1234", "host_name":"test.icalendar"},),
    ("-", "abc@test.example.com", ("test.example.com", "abc"), {},),

])
def test_uid_generator_issue_345(args, kw, split, expected):
    '''Issue #345 - Why is tools.UIDGenerator a class (that must be instantiated) instead of a module?

    see https://github.com/collective/icalendar/issues/345
    '''
    uid = UIDGenerator.uid(*args, **kw)
    assert uid.split(split)[1] == expected



"""These are tests for Issue #348

see https://github.com/collective/icalendar/issues/348
"""


def test_calendar_can_be_parsed_correctly(calendars):
    """Exception when there's no ':' when parsing value #348

    see https://github.com/collective/icalendar/issues/348 
    """
    freebusy = calendars.issue_348_exception_parsing_value.walk("VFREEBUSY")[0]
    assert freebusy["ORGANIZER"].params["CN"] == "Sixt SE"


def test_parameters_are_not_truncated(calendars):
    """We skip to the end and we do not want to loose parameters.

    see https://github.com/collective/icalendar/pull/514#issuecomment-1505878801
    """
    freebusy = calendars.issue_348_exception_parsing_value.walk("VFREEBUSY")[0]
    assert freebusy["X-ORGANIZER2"].params["CN"] == "Sixt SE"
    assert freebusy["X-ORGANIZER2"].params["CN2"] == "Test!"



#!/usr/bin/env python3
"""utility program that allows user to preview calendar's events"""
import sys
import pathlib
import argparse
from datetime import datetime

from icalendar import Calendar, __version__

def _format_name(address):
    """Retrieve the e-mail and the name from an address.

    :arg an address object, e.g. mailto:test@test.test

    :returns str: The name and the e-mail address.
    """
    email = address.split(':')[-1]
    name = email.split('@')[0]
    if not email:
        return ''
    return f"{name} <{email}>"


def _format_attendees(attendees):
    """Format the list of attendees.

    :arg any attendees: Either a list, a string or a vCalAddress object.

    :returns str: Formatted list of attendees.
    """
    if isinstance(attendees, list):
        return '\n'.join(map(lambda s: s.rjust(len(s) + 5), map(_format_name, attendees)))
    return _format_name(attendees)

def view(event):
    """Make a human readable summary of an iCalendar file.

    :returns str: Human readable summary.
    """
    summary = event.get('summary', default='')
    organizer = _format_name(event.get('organizer', default=''))
    attendees = _format_attendees(event.get('attendee', default=[]))
    location = event.get('location', default='')
    comment = event.get('comment', '')
    description = event.get('description', '').split('\n')
    description = '\n'.join(map(lambda s: s.rjust(len(s) + 5), description))

    start = event.decoded('dtstart')
    if 'duration' in event:
        end = event.decoded('dtend', default=start + event.decoded('duration'))
    else:
        end = event.decoded('dtend', default=start)
    duration = event.decoded('duration', default=end - start)
    if isinstance(start, datetime):
        start = start.astimezone()
    start = start.strftime('%c')
    if isinstance(end, datetime):
        end = end.astimezone()
    end = end.strftime('%c')

    return f"""    Organizer: {organizer}
    Attendees:
{attendees}
    Summary    : {summary}
    Starts     : {start}
    End        : {end}
    Duration   : {duration}
    Location   : {location}
    Comment    : {comment}
    Description:
{description}"""

def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('calendar_files', nargs='+', type=pathlib.Path)
    parser.add_argument('--output', '-o', type=argparse.FileType('w'), default=sys.stdout, help='output file')
    parser.add_argument('-v', '--version', action='version', version=f'{parser.prog} version {__version__}')
    argv = parser.parse_args()

    for calendar_file in argv.calendar_files:
        with open(calendar_file) as f:
            calendar = Calendar.from_ical(f.read())
            for event in calendar.walk('vevent'):
                argv.output.write(view(event) + '\n\n')

if __name__ == '__main__':
    main()



from icalendar import Event
from datetime import date, datetime

import pytest

def test_recurrence_properly_parsed(events):
    assert events.event_with_recurrence['rrule'] == {'COUNT': [100], 'FREQ': ['DAILY']}

@pytest.mark.parametrize('i, exception_date', [
    (0, datetime(1996, 4, 2, 1, 0)),
    (1, datetime(1996, 4, 3, 1, 0)),
    (2, datetime(1996, 4, 4, 1, 0))
])
def test_exdate_properly_parsed(events, i, exception_date, in_timezone):
    assert events.event_with_recurrence['exdate'].dts[i].dt == in_timezone(exception_date, 'UTC')

def test_exdate_properly_marshalled(events):
    actual = events.event_with_recurrence['exdate'].to_ical()
    assert actual == b'19960402T010000Z,19960403T010000Z,19960404T010000Z'

# TODO: DOCUMENT BETTER!
# In this case we have multiple EXDATE definitions, one per line.
# Icalendar makes a list out of this instead of zipping it into one
# vDDDLists object. Actually, this feels correct for me, as it also
# allows to define different timezones per exdate line - but client
# code has to handle this as list and not blindly expecting to be able
# to call event['EXDATE'].to_ical() on it:
def test_exdate_formed_from_exdates_on_multiple_lines_is_a_list(events):
    exdate = events.event_with_recurrence_exdates_on_different_lines['exdate']
    assert isinstance(exdate, list)

@pytest.mark.parametrize('i, exception_date, exception_date_ics', [
    (0, datetime(2012, 5, 29, 10, 0), b'20120529T100000'),
    (1, datetime(2012, 4, 3, 10, 0),  b'20120403T100000'),
    (2, datetime(2012, 4, 10, 10, 0), b'20120410T100000'),
    (3, datetime(2012, 5, 1, 10, 0),  b'20120501T100000'),
    (4, datetime(2012, 4, 17, 10, 0), b'20120417T100000')
])
def test_list_exdate_to_ical_is_inverse_of_from_ical(events, i, exception_date, exception_date_ics, in_timezone):
    exdate = events.event_with_recurrence_exdates_on_different_lines['exdate']
    assert exdate[i].dts[0].dt == in_timezone(exception_date, 'Europe/Vienna')
    assert exdate[i].to_ical() == exception_date_ics

@pytest.mark.parametrize('freq, byday, dtstart, expected', [
    # Test some YEARLY BYDAY repeats
    ('YEARLY', '1SU', date(2016,1,3), # 1st Sunday in year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 1SU\r\nDTSTART;VALUE=DATE:20160103\r\nRRULE:FREQ=YEARLY;BYDAY=1SU\r\nEND:VEVENT\r\n'),
    ('YEARLY', '53MO', date(1984,12,31), # 53rd Monday in (leap) year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 53MO\r\nDTSTART;VALUE=DATE:19841231\r\nRRULE:FREQ=YEARLY;BYDAY=53MO\r\nEND:VEVENT\r\n'),
    ('YEARLY', '-1TU', date(1999,12,28), # Last Tuesday in year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY -1TU\r\nDTSTART;VALUE=DATE:19991228\r\nRRULE:FREQ=YEARLY;BYDAY=-1TU\r\nEND:VEVENT\r\n'),
    ('YEARLY', '-17WE', date(2000,9,6), # 17th-to-last Wednesday in year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY -17WE\r\nDTSTART;VALUE=DATE:20000906\r\nRRULE:FREQ=YEARLY;BYDAY=-17WE\r\nEND:VEVENT\r\n'),
    # Test some MONTHLY BYDAY repeats
    ('MONTHLY', '2TH', date(2003,4,10), # 2nd Thursday in month
        b'BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY 2TH\r\nDTSTART;VALUE=DATE:20030410\r\nRRULE:FREQ=MONTHLY;BYDAY=2TH\r\nEND:VEVENT\r\n'),
    ('MONTHLY', '-3FR', date(2017,5,12), # 3rd-to-last Friday in month
        b'BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY -3FR\r\nDTSTART;VALUE=DATE:20170512\r\nRRULE:FREQ=MONTHLY;BYDAY=-3FR\r\nEND:VEVENT\r\n'),
    ('MONTHLY', '-5SA', date(2053,11,1), # 5th-to-last Saturday in month
        b'BEGIN:VEVENT\r\nSUMMARY:Event MONTHLY -5SA\r\nDTSTART;VALUE=DATE:20531101\r\nRRULE:FREQ=MONTHLY;BYDAY=-5SA\r\nEND:VEVENT\r\n'),
    # Specifically test examples from the report of Issue #518
    # https://github.com/collective/icalendar/issues/518
    ('YEARLY', '9MO', date(2023,2,27), # 9th Monday in year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 9MO\r\nDTSTART;VALUE=DATE:20230227\r\nRRULE:FREQ=YEARLY;BYDAY=9MO\r\nEND:VEVENT\r\n'),
    ('YEARLY', '10MO', date(2023,3,6), # 10th Monday in year
        b'BEGIN:VEVENT\r\nSUMMARY:Event YEARLY 10MO\r\nDTSTART;VALUE=DATE:20230306\r\nRRULE:FREQ=YEARLY;BYDAY=10MO\r\nEND:VEVENT\r\n'),
])
def test_byday_to_ical(freq, byday, dtstart, expected):
    'Test the BYDAY rule is correctly processed by to_ical().'
    event = Event()
    event.add('SUMMARY', ' '.join(['Event', freq, byday]))
    event.add('DTSTART', dtstart)
    event.add('RRULE', {'FREQ':[freq], 'BYDAY':byday})
    assert event.to_ical() == expected



import codecs
import setuptools
import re
import ast

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('src/icalendar/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))


shortdesc = 'iCalendar parser/generator'
longdesc = ''
for fname in ['README.rst', 'CONTRIBUTING.rst', 'CHANGES.rst', 'LICENSE.rst']:
    with codecs.open(fname, encoding='utf-8') as f:
        longdesc += f.read()
        longdesc += '\n'

tests_require = []
install_requires = [
    'python-dateutil',
    'pytz',
    # install requirements depending on python version
    # see https://www.python.org/dev/peps/pep-0508/#environment-markers
    'backports.zoneinfo; python_version == "3.7" or python_version == "3.8"',
]


setuptools.setup(
    name='icalendar',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='calendar calendaring ical icalendar event todo journal '
             'recurring',
    author='Plone Foundation',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://github.com/collective/icalendar',
    license='BSD',
    packages=setuptools.find_namespace_packages('src', exclude=["icalendar.fuzzing"]),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=install_requires,
    entry_points = {'console_scripts': ['icalendar = icalendar.cli:main']},
    extras_require={
        'test': tests_require
    },
    test_suite='icalendar.tests'
)
